%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright 2013-2014. All rights reserved.                               %
% Institute of Measurement and Control Systems                            %
% Karlsruhe Institute of Technology, Germany                              %
%                                                                         %
% This file is part of triTrack2.                                         %
% Authors:  Philip Lenz                                                   %
%           Please send any bugreports to lenz@kit.edu                    % 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function triTrack2(tritrack_user, matcher_user, display_user)
% TRITRACK2 main function for sparse scene flow segmentation object tracking

% allows function to be called as a script
stack=dbstack();
if strcmp(stack(end).name, 'triTrack2')
  close all; clc; dbstop error;
end
global key close_requested;
key             = '';
close_requested = false;
addpath('./src');
addpath('./bgl');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SETTINGS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% empty if not provided
if nargin<3, display_user=struct(); end
if nargin<2, matcher_user=struct(); end
if nargin<1, tritrack_user=struct(); end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SPECIFY DISPLAY OPTIONS
global display;
% If you encounter issues like that one:
% 'java.lang.IllegalArgumentException: adding a container to a container on a different GraphicsDevice'
% enable this switch which de-activates transparent drawing
display.nvidia_workaround = true;
% select camera to use (impact on matcher and VO possible)
display.cam = 2;
% show result and precessing information on-line
display.flush             = true;  % update drawing after every iteration
display.stop_at_end       = true;  % behaviour at the end of a sequence
display.console.verbose   = false; % print status messages
display.console.time      = false; % print timing in console
% specify what should be shown in the image
display.image.figure            = true;
display.image.staticFeatures    = false; % show static features as small red dots
display.image.untrackedFeatures = false; % show untracked features colored in their classiciation
display.image.trackedFeatures   = false; % show untracked features colored in their classiciation
display.image.patches           = false; % show subgraphs as patches
display.image.bbox              = true; % show the bounding box of a detected object
display.image.ID                = true; % show tracking ID for each object
display.image.trajectory        = true; % show the trajectory (measurements and filter)
display.image.prediction        = false; % show the prediction of th trajectory 
% specify what should be show in world coordinates
display.wcs.figure  = true;
display.wcs.FOV     = false;  % show field of view
% specify what should be shown in image coordinates (useful for debugging)
display.ics.figure                  = true;
display.ics.untrackedFeatures       = false; % show untracked features colored in their classiciation
display.ics.trackedFeatures         = false; % show tracked features colored in their classiciation
display.ics.trackedFeaturesVelocity = false; % show scene flow for tracked features
% save visualization
display.image.save    = true;
display.ics.save      = true;
display.wcs.save      = true;
% merge
display = set_user_params(display,display_user);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SPECIFY SEQUENCE AND ALGORITHM PARAMETERS

% merge tritrack defaults and user parameter
tritrack_params.base_path     = '/media/PHILIP_KITTI/datasets/kitti/2012_tracking/training';
tritrack_params.seq_idx       = 8;
tritrack_params.start_frame   = 0;
tritrack_params.end_frame     = 0;
% save
tritrack_params.cache           = false;
tritrack_params.save_detections = false;
% detection
tritrack_params.mahal_dist    = 1.2;
tritrack_params.match_dist    = 1.8;
tritrack_params.min_occurence = 5;
% system parameters
tritrack_params.deltaT = 1/10; % cameras are triggered to 10Hz velodyne
% thresholds
tritrack_params.thres.maxDist      = 90;
tritrack_params.thres.maxLat       = 30;
tritrack_params.thres.d_min        = 10 * tritrack_params.deltaT;
tritrack_params.thres.V_min        = 1.5 * tritrack_params.deltaT;
tritrack_params.thres.V_max        = 100 * tritrack_params.deltaT;
% association parameter
tritrack_params.ass.maxTracking  = 3;
tritrack_params.ass.maxDist      = 2;
tritrack_params.ass.maxVelocity  = 0.15;
% merge
tritrack_params = set_user_params(tritrack_params,tritrack_user);
if isfield(tritrack_params, 'drive') && isfield(tritrack_params, 'date')
  tritrack_params.results_path  = fullfile(tritrack_params.base_path,'tritrack',...
    sprintf('%s_drive_%04d_%s',tritrack_params.date,tritrack_params.drive,datestr(now,'YYYY_MM_DD-hh_mm_ss')));
else
  tritrack_params.results_path  = fullfile(tritrack_params.base_path,'tritrack',...
    sprintf('%04d_%s',tritrack_params.seq_idx,datestr(now,'YYYY_MM_DD-hh_mm_ss')));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SPECIFY MATCHER AND VISUAL ODOMETRY PARAMETER
matcher_params.nms_n                  = 5;   % non-max-suppression: min. distance between maxima (in pixels)
matcher_params.nms_tau                = 50;  % non-max-suppression: interest point peakiness threshold
matcher_params.match_binsize          = 50;  % matching bin width/height (affects efficiency only)
matcher_params.match_radius           = 200; % matching radius (du/dv in pixels)
matcher_params.match_disp_tolerance   = 1;   % du tolerance for stereo matches (in pixels)
matcher_params.outlier_disp_tolerance = 5;   % outlier removal: disparity tolerance (in pixels)
matcher_params.outlier_flow_tolerance = 5;   % outlier removal: flow tolerance (in pixels)
matcher_params.multi_stage            = 1;   % 0=disabled,1=multistage matching (denser and faster)
matcher_params.half_resolution        = 0;   % 0=disabled,1=match at half resolution, refine at full resolution
matcher_params.refinement             = 2;   % refinement (0=none,1=pixel,2=subpixel)
% bucketing parameters
matcher_params.max_features         = 4;
matcher_params.bucket_width         = 10;
matcher_params.bucket_height        = 10;
% merge
matcher_params = set_user_params(matcher_params,matcher_user);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SET PATHS AND CREATE NEEDED DIRECTORIES

% if field "drive" is provided, triTrack uses a KITTI raw sequence
if isfield(tritrack_params, 'drive') && isfield(tritrack_params, 'date')
  imgL_file    = fullfile(tritrack_params.base_path,...
                          tritrack_params.date,...
                          sprintf('%s_drive_%04d_sync', tritrack_params.date, tritrack_params.drive),...
                          sprintf('image_%02d/data',display.cam),'%010d.png');
  imgR_file    = fullfile(tritrack_params.base_path,...
                          tritrack_params.date,...
                          sprintf('%s_drive_%04d_sync', tritrack_params.date, tritrack_params.drive),...
                          sprintf('image_%02d/data', display.cam+1),'%010d.png');
  matches_path = fullfile(tritrack_params.base_path,...
                          tritrack_params.date,...
                          sprintf('%s_drive_%04d', tritrack_params.date, tritrack_params.drive),...
                          'quadmatches');
  calib_dir = fullfile(tritrack_params.base_path,tritrack_params.date);
% otherwise the tracking benchmark is used
else
  imgL_file    = fullfile(tritrack_params.base_path,sprintf('image_%02d',display.cam),...
                          sprintf('%04d',tritrack_params.seq_idx), '%06d.png');
  imgR_file    = fullfile(tritrack_params.base_path,sprintf('image_%02d',display.cam+1),...
                          sprintf('%04d',tritrack_params.seq_idx), '%06d.png');
  matches_path = fullfile(tritrack_params.base_path,'quadmatches',...
                          sprintf('%04d',tritrack_params.seq_idx));
  calib_dir = fullfile(tritrack_params.base_path,'calib');
end
  % create directories for cache and results, if needed
  if tritrack_params.save_detections 
    tritrack_params.detections_folder = fullfile(fileparts(tritrack_params.results_path), 'det_02');
    if ~exist(tritrack_params.detections_folder,'dir')
      try
        mkdir(tritrack_params.detections_folder);
      catch
        fprintf(2, '  . Could not access result path. Detections won''t be saved.\n');
        tritrack_params.save_detections = false;
      end
    end
  end
  if(~exist(tritrack_params.results_path,'dir'))
    try
      if any([display.image.save, display.ics.save, display.wcs.save])
        mkdir(tritrack_params.results_path);
      end
    catch %#ok<CTCH>
      fprintf(2, '  . Could not access result path. Results will not be produced.\n');
      display.image.save = false;
      display.ics.save = false;
      display.wcs.save = false;
    end
    if display.image.save && ~exist(fullfile(tritrack_params.results_path,'img'),'dir')
      mkdir(fullfile(tritrack_params.results_path,'img'));
    end
    if display.ics.save && ~exist(fullfile(tritrack_params.results_path,'ics'),'dir')
      mkdir(fullfile(tritrack_params.results_path,'ics'));
    end
    if display.wcs.save && ~exist(fullfile(tritrack_params.results_path,'wcs'),'dir')
      mkdir(fullfile(tritrack_params.results_path,'wcs'));
    end
  end
  if(~exist(matches_path,'dir') && tritrack_params.cache)
    try
      mkdir(matches_path);
    catch %#ok<CTCH>
      fprintf(2, '  . Could not access cache path. No matches will be cached\n');
      tritrack_params.cache = false;
    end
  end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LOAD CALIBRATION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

try
  cam                    = readCalibration(calib_dir,tritrack_params.seq_idx, display.cam);
  cam.mounting_pitch_deg = 0;   % this must unforunately be set by hand right now
  cam.height_over_ground = 1.65; % this must unforunately be set by hand right now
catch  %#ok<CTCH>
  fprintf(2, '  . Could not find calibration file. Exiting...\n');
  return
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INIT LOOP
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% MATCHER
matcherMex('init',matcher_params);
% VISUAL ODOMETRY
odo_params      = cam;
odo_params.base = -odo_params.base;
visualOdometryStereoMex('init',odo_params);
% VO-POSES
Tr_total{1} = eye(4);
% TRACKS
newID           = 1;
track           = {};
last_obj        = {};
acc.active      = false;
acc.track       = [];
% VISUALIZATION
if ~exist(sprintf(imgL_file, tritrack_params.start_frame),'file')
  % fallback for jpg image data
  imgL_file    = fullfile(tritrack_params.base_path,sprintf('image_%02d',display.cam),...
                          sprintf('%04d',tritrack_params.seq_idx), '%06d.jpg');
  imgR_file    = fullfile(tritrack_params.base_path,sprintf('image_%02d',display.cam+1),...
                          sprintf('%04d',tritrack_params.seq_idx), '%06d.jpg');
  if ~exist(sprintf(imgL_file, tritrack_params.start_frame),'file')
    fprintf(2, '  . Image data not found. Exiting...\n');
    return
  end
end
h    = visualization('init', sprintf(imgL_file,tritrack_params.start_frame), cam, tritrack_params.thres);
% DE-ACTIVATE KNOWN WARNINGS
warning('off','MATLAB:DelaunayTri:DupPtsWarnId');
% START FRAME
frame      = tritrack_params.start_frame; % frames in KITTI are zero-indexed
last_frame = nan;
play       = true;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MAIN LOOP
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  finishup = onCleanup(@() cleanup()); % there is no way of really catching ctrl-c :-(
  while 1

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % TINY UI AND INIT
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % get keytroke asynchronous
    for hh=1:numel(h)
      try
        set(h{hh}.fig,'CloseRequestFcn',@close_request_fcn);
        if close_requested, close all force; return; end
        set(h{hh}.fig,'KeyPressFcn',@key_press_fcn);
      catch %#ok<CTCH>
        continue
      end
    end
    % evaluate key
    switch lower(key)
      case 'space'
        play = ~play;
      case 'add'
        frame = increment_frame(imgL_file, frame);
      case 'a'
        active_track_present = false;
        for itTrack=1:numel(track)
          if track{itTrack}.active
            active_track_present = true;
            break;
          end
        end
        if ~acc.active && ~active_track_present
          fprintf(2,'  . No tracks existing. Cannot switch to ACC mode.\n');
        else
          acc.active = ~acc.active;
          if acc.active
            try %#ok<*TRYNC>
              selection = input('Select Track ID for ACC Mode: ');
              if track{selection}.active
                acc.track = selection;
              end
            end
          end
        end
      case {'q', 'escape'}
        close all force;
        return;
    end
    key = '';
    if play
      frame = increment_frame(imgL_file, frame); 
    end
    if last_frame==frame && play;
      if ~display.stop_at_end
        close all force; 
        return;
      else
        pause(0.1);
        continue;
      end
    end

    % init iteration
    time = 0;

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % MATCHING
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % read current images
    % frames in KITTI are zero-indexed, frame is incremented in the beginning
    % of the loop
    imgL = sprintf(imgL_file,frame-1);
    imgR = sprintf(imgR_file,frame-1);
    if ~exist(imgL,'file'), break; end
    if display.console.verbose, fprintf('. Process frame no. %d\n',frame-tritrack_params.start_frame-1); end
    tic;
    I1 = imread(imgL);

    % TODO: Check whole cache and only load, if everything exists
    cache_matchfile = fullfile(matches_path, sprintf('%06d.mat',frame-1));
    if exist(cache_matchfile,'file')
      if(display.console.time), fprintf('   . reading images in %f seconds\n',toc); end
      time=time+toc; tic;
      load(cache_matchfile, 'p_matched','Tr','Tr_total');
      if(display.console.time), fprintf('   . loaded cached matching data in %f seconds\n',toc); end
      time=time+toc; tic;
    else
      I2 = imread(imgR);
      if(display.console.time), fprintf('   . reading images in %f seconds\n',toc); end
      time=time+toc;tic;
      % egomotion
      % estimate egomotion, failure+1 is the number of frames between matches
      if display.cam>1
        Tr = visualOdometryStereoMex('process',rgb2gray(I1),rgb2gray(I2));
      else
        Tr = visualOdometryStereoMex('process',I1,I2);
      end
      % accumulate total transformation
      Tr_total{frame-tritrack_params.start_frame+1} = Tr_total{frame-tritrack_params.start_frame}*Tr^-1; %#ok<*AGROW>
      if(display.console.time), fprintf('   . visual odometry took %f seconds\n',toc); end
      time=time+toc; tic;
      % matches
      if display.cam>1
        matcherMex('push',rgb2gray(I1),rgb2gray(I2));
      else
        matcherMex('push',I1,I2);
      end
      if(frame==tritrack_params.start_frame+1), continue; end
      matcherMex('match',2);
      p_matched = matcherMex('get_matches',2);
      if(display.console.time), fprintf('   . matching took %f seconds (%d features detected)\n',toc, size(p_matched,2)); end
      time=time+toc;tic;
      if tritrack_params.cache
        save(cache_matchfile,'p_matched','Tr','Tr_total');
      end
    end

    % get 3D coordinates
    % current time step
    tic;
    [Xc, Yc, Zc] = ...
      projectImgTo3d(p_matched(5,:), p_matched(7,:), p_matched(6,:), cam);
    % previous time step
    [Xp, Yp, Zp] = ...
      projectImgTo3d(p_matched(1,:), p_matched(3,:), p_matched(2,:), cam);
    % Egomotion coordinate system is Y pointing down.
    Yc=-Yc;
    Yp=-Yp;
    % corrrect egomotion
    % project previous 3D coordinates in current coordinate system
    % using homogeneous coordinates [X, Y, Z, 1]' and Transformation Matrix Tr
    if ~isempty(Xp) && ~isempty(Xc)
      tmp=Tr*[Xp; Yp; Zp;ones(1,size(Xp,2))];
      Xp=tmp(1,:);
      Yp=tmp(2,:);
      Zp=tmp(3,:);
      % handle numerical problems (only consider reasonable range)
      delInd = abs(Xc)>tritrack_params.thres.maxLat | abs(Zc)>tritrack_params.thres.maxDist | abs(Xp)>tritrack_params.thres.maxLat | abs(Zp)>tritrack_params.thres.maxDist;
      Xp(delInd)=[]; Yc(delInd)=[]; Zp(delInd)=[]; Xc(delInd)=[]; Yp(delInd)=[]; Zc(delInd)=[];
      p_matched(:,delInd)=[];
      if(display.console.time), fprintf('   . 3D reconstruction took %f seconds\n',toc); end
      time=time+toc;
      % compute translation (motion*tritrack_params.deltaT) of matches
      tic;
      [tx,ty,tz,J] = computeTranslation({[Xp;Xc],[Yp;Yc],[Zp;Zc]},{[p_matched(1,:);...
        p_matched(5,:)], [p_matched(3,:); p_matched(7,:)], [p_matched(2,:); p_matched(4,:)],},cam);
      if(display.console.time), fprintf('   . velocity computation took %f seconds\n',toc); end
      time=time+toc;
    else
      delInd = [];
      tx = []; ty = []; tz = []; J = {};
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % classify unkown, dynamic and static matches
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    if ~isempty(tx)
      tic;
      V_abs = sqrt(tx.^2+tz.^2);
      d = zeros(1,numel(J));
      for i=1:numel(J)
        thisJ = J{i}*J{i}';
        thisJ = [thisJ(1,1),thisJ(1,3); thisJ(3,1), thisJ(3,3)];
        d(i) = myMahal(thisJ^(-1),[tx(i),tz(i)]');
      end

      % point close to the optical axis
      indUnknown = ((Xc<2 & Xc >-5) & abs(Zc)>10) & V_abs>0.5;

      % definitly static points
      indStatic = ((d<tritrack_params.thres.d_min | ...
                    V_abs<tritrack_params.thres.V_min) & ...
                   V_abs<tritrack_params.thres.V_max & ~indUnknown);

      if(display.console.time), fprintf('   . egomotion compensation took %f seconds\n',toc); end
      time=time+toc;
    else
      V_abs = []; d = []; indUnknown = []; indStatic = [];
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % GRAPH SMOOTHING
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    if ~isempty(Xc)
      tic;
      [indSingularity,indUnknown] = graphSmoothing(Xc, Zc, indStatic, indUnknown, p_matched);
      if(display.console.time), fprintf('   . graph smoothing took %f seconds\n',toc); end
      time=time+toc;
      % set dynamic indices
      ind_d = ( (~indStatic & ~indSingularity) | indUnknown);
    else
      indSingularity = []; indUnknown = []; ind_d = [];
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % OBJECT DETECTION
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    tic;
    obj = getObjects(tritrack_params,Xc, Yc, Zc, tx, ty, tz, J, indStatic, indSingularity, indUnknown, cam);
    [obj, track, newID] = associateObjects(last_obj, obj, track, newID,...
      tritrack_params.min_occurence, Tr_total, frame, cam);
    [track, obj] = associateTracklets(track, obj, Tr_total, [Xc;Yc;Zc], [tx;ty;tz], frame,...
      tritrack_params.ass.maxTracking, tritrack_params.ass.maxDist,...
      tritrack_params.ass.maxVelocity, display.console.verbose);
    track = mergeTracks(track, Tr_total{end}, display.console.verbose);
    if(display.console.time), fprintf('   . object detection took %f seconds\n',toc); end
    time=time+toc;

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % VISUALIZATION
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    tic;
    % draw objects
    [h, track] = visualization('draw', h, I1, frame,...
      ind_d, tritrack_params.min_occurence, tritrack_params.deltaT, tritrack_params.thres, Tr_total, p_matched, obj, track, acc, cam);
    if(display.console.time), fprintf('   . visualization took %f seconds\n',toc); end
    time=time+toc;

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % SAVE DATA
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    tic
    if tritrack_params.save_detections
      saveRawObjects(frame,obj,tritrack_params,cam,Tr_total{end});
    end
    if(display.image.figure   &&   display.image.save)
      % update usage and frame number for video
      delete(h{1}.usage, h{1}.frame);
      % set(h{1}.frame,'fontsize',18);
      img_file = fullfile(fullfile(tritrack_params.results_path,'img',sprintf('%06d.jpg',frame-1)));
      saveas(h{1}.fig, img_file, 'jpg')
    end
    if(display.wcs.figure   &&   display.ics.save)
      ICS_file = fullfile(fullfile(tritrack_params.results_path,'ics',sprintf('%06d.jpg',frame-1)));
      saveas(h{2}.fig, ICS_file, 'jpg')
    end
    if(display.ics.figure   &&   display.wcs.save)
      WCS_file = fullfile(fullfile(tritrack_params.results_path,'wcs',sprintf('%06d.jpg',frame-1)));
      saveas(h{3}.fig, WCS_file, 'jpg')
    end
    if(display.console.time && any([display.wcs.save,display.ics.save,display.image.save]))
      fprintf('   . saving images took %f seconds\n',toc);
    end
    time = time+toc;
    if display.console.verbose, fprintf('   . frame processing took %f seconds\n',time); end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % FINISH ITERATION
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    last_obj   = obj;
    last_frame = frame;

  end % while

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FINISH
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

matcherMex('close');
visualOdometryStereoMex('close');
if display.console.verbose, fprintf('.  ALL DONE !\n'); end
end %function plotEgomotion

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% HELPER FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function params = set_user_params(params, user)
% SET_USER_PARAMS sets (for structured strucs recursively) the parameters 
%    specified in 'user'
  fields = fieldnames(user);
  for it=1:numel(fields)
    if isstruct(getfield(user,fields{it})) && any(strcmp(fields{it},fieldnames(user))) %#ok<*GFLD>
      params = setfield(params,fields{it}, set_user_params(getfield(params,fields{it}), getfield(user,fields{it}))); %#ok<*SFLD>
    else
      params = setfield(params,fields{it},getfield(user,fields{it})); %#ok<GFLD,SFLD>
    end
  end
end

function [frame] = increment_frame(imgL_file, frame, increment)
  if ~exist('increment','var'), increment=1; end
  new_frame = frame+increment;
  imgL = sprintf(imgL_file,new_frame);
  if exist(imgL,'file'), frame=new_frame; end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CALLBACKS
function key_press_fcn(h,e) %#ok<*INUSL>
  global key
  key = e.Key;
end

function close_request_fcn(h,e) %#ok<INUSD>
  global close_requested
  close_requested = true;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CLEANUP
function cleanup()
dbclear error
close all force
disp('Caught CTRL-C. Exiting...')
end
